/*
 * Copyright (C) 2013 by the University of Southern California
 * $Id: dns_query.cc 19925 Tue Feb 11 14:41:03 PST 2014 Zi Hu $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/*
   t-DNS client side proxy.
   by Zi Hu (zihu@usc.edu) 2013-09-18
*/

#include<stdio.h>
#include<string.h>    //strlen
#include<stdlib.h>    
#include<sys/socket.h>  
#include<arpa/inet.h> //inet_addr , inet_ntoa , ntohs etc
#include<netinet/in.h>
#include<pthread.h>
#include<signal.h>
#include<unistd.h>
#include "utility.h"
#include "types.h" 
#include "mysocket.h"
#include <sys/ioctl.h>
#include <linux/sockios.h> /* SIOCOUTQ */
//#include <chrono>
#include <ctime>
#include <getopt.h>
#include <sys/types.h>
#include <unistd.h>
#include "ssl_util.h"

#define DISABLE_NAGLE false

#define R_SERV "128.9.168.212"

#define CAFILE "/etc/ssl/certs/ca-bundle.trust.crt"


using namespace std;

//thread to forward query from local to remote
pthread_t td_local_to_remote;
//thread to forward response from remote to local 
pthread_t td_remote_to_local;

// remote Server ip
char dns_servers[MAX_BUF_SIZE];
//local ip

//remote server port
unsigned short dns_port = 10053;
//local port to listen on
unsigned short local_port = 53;

//disable nagle?
bool disable_nagle=false;

//send queries over tls directly (without tls negotiation)
bool direct_tls = false;

bool SSL_init_done = false;

//enable TCP fastopen?
bool enable_tcpfast=false;
//validate server certificate?
bool validate_ca = false;

bool first_packet = true;

//which protocol
string proto = "tcp";

//retry connection timeout
int retry_timeout = 1;
//number of retries
int retry_count = 15;


//log file
FILE * log_fp;

struct conn_hub {
  CSocket* udp_conn;
  CSocket* tcp_conn;
  SSL* ssl_conn;
  string conn_type;
};


struct conn_hub* conn_point = new struct conn_hub;


void clean_old_connection(struct conn_hub *conn_p);
void cleanup();



//set up ssl handshake
SSL* ssl_handshake(CSocket *mysock)
{

  SSL_CTX* ctx = SSL_CTX_new (SSLv23_client_method ());
  if (!ctx) ERR_print_errors_fp (log_fp);
  SSL *ssl = new SSL;
  ssl = (SSL*)outgoing_ssl_fd(ctx, mysock->_sock);

  if(!ssl) 
  {
    fprintf(log_fp, "cannot create ssl\n");
    return NULL;
  }

  while(1) 
  {
    int r;
    ERR_clear_error();
    if( (r=SSL_do_handshake(ssl)) == 1)
      break;

    r = SSL_get_error(ssl, r); 
    if(r != SSL_ERROR_WANT_READ && r != SSL_ERROR_WANT_WRITE) 
    {
      fprintf(log_fp, "could not ssl_handshake\n");
      return NULL;
    }   
  }   


  if(validate_ca) 
  {
    SSL_CTX_load_verify_locations(ctx, CAFILE, NULL);
    X509* x = SSL_get_peer_certificate(ssl);
    if(!x) 
    {    
      fprintf(log_fp, "SSL: no peer certificate\n");
    }    
    else 
    {    
      //X509_print_fp(stdout, x); 
      long res=SSL_get_verify_result(ssl);
      if (res == X509_V_OK)
      {
        printf("validate success:%ld\n", res);
        X509_free(x);
      }
      else
      {
        printf("validate fail:%ld\n", res);
        X509_free(x);
        exit(0);
      }
    }    
  }
  return ssl;
}


//receive response on TCP connection
int recv_tcp_response(CSocket *mysock, char* recv_buf, int buf_size)
{
  char buf[2];
  memset(buf, 0, 2);
  char pkt_len_buf[2];
  memset(pkt_len_buf, 0, 2);
  char pkt_buf[MAX_BUF_SIZE];
  memset(pkt_buf, 0, MAX_BUF_SIZE);

  unsigned short pkt_len=0;
  int toread = 0;
  int recv_bytes =0;
  //read first two bytes (the TCP DNS packet length);
  memset(pkt_len_buf, 0, 2);
  while(toread<2)
  {
    recv_bytes=mysock->recv_data_TCP((char*)buf, 1);
    if(recv_bytes<=0)
    {
      printf("receive bytes: %d\n", recv_bytes);
      return -1;
    }

    if(recv_bytes==1)
    {
      pkt_len_buf[toread]=buf[0];
      toread++;
    }
  }
  pkt_len = ntohs(*((unsigned short*) pkt_len_buf));

  if(pkt_len+2 > buf_size)
  {
    fprintf(log_fp, "buf is not big enough to store the whole the tcp response");
    return -1;
  }
  memcpy(recv_buf, pkt_len_buf, 2);

  //read the rest packet;
  toread=0; 
  memset(pkt_buf, 0, MAX_BUF_SIZE);
  while(toread<pkt_len)
  {
    recv_bytes=mysock->recv_data_TCP((char*)buf, 1);
    if(recv_bytes<=0)
    {
      //printf("receive bytes: %d\n", recv_bytes);
      return -1;
    }

    if(recv_bytes==1)
    {
      pkt_buf[toread]=buf[0];
      toread++;
    }
  }
  memcpy(recv_buf+2, pkt_buf, pkt_len);
  return pkt_len+2;
}

// send dummy query to see if resolver supports TLS. 
bool resolver_tls_check(CSocket &mysock)
{
    //printf("perform TLS negotiation\n");
    unsigned char buf[MAX_BUF_SIZE];
    memset(buf, 0, MAX_BUF_SIZE);

    unsigned short plen;
 
    int splen=0;
    int recv_len=0;
 

    //construct a fake EDNS query packet; 
    //query "." with EDNS query (with the second bit in "Z" set, the bit right after the "DO" bit), 
    //plen = construct_dns_query_EDNS(buf, MAX_BUF_SIZE, (char*)".", ns_c_in, ns_t_a, true);
    plen = construct_dns_query_EDNS(buf, MAX_BUF_SIZE, (char*)"STARTTLS", ns_c_chaos, ns_t_txt, true);


    if(enable_tcpfast && first_packet)
    {
      struct sockaddr_in dest;
      dest.sin_family = AF_INET;
      dest.sin_port = htons(dns_port);
      dest.sin_addr.s_addr = inet_addr(dns_servers); //dns servers

      //send the query with TCP fastopen. 
      if((splen=mysock.send_data_TCP_fo((char*)buf, plen, dest))<0)
      {
        exit(0);
      }

      first_packet=false;
    }
    else
    {
      //send the query via normal TCP.
      if((splen=mysock.send_data_TCP((char*)buf, plen))<0)
      {
        exit(0);
      }
    }

    //send the query 
    //if((splen=mysock.send_data_TCP((char*)buf, plen))<0)
    //{
    //  exit(0);
    //}


    //Receive the answer
    memset(buf, 0, MAX_BUF_SIZE);
    if((recv_len=mysock.recv_data_TCP((char*)buf))<0)
    //if((recv_len=recv_tcp_response(&mysock,(char*)buf, MAX_BUF_SIZE)) <0 )
    {
      fprintf(log_fp, "receive error:%d\n", recv_len);
      exit(0);
    }

    //parse the dns response;
    //parse_dns_response((unsigned char*)buf+2, recv_len-2);
    bool tls_bit_set = is_tls_set(buf+2, recv_len-2);
    if(tls_bit_set)
    {
      printf("remote server support DNS-over-TLS\n");
      return true;
    }
    else
    {
      return false;
    }
}


//receive response on SSL.
int recv_ssl_response(SSL* myssl, char* recv_buf, int buf_size)
{
  char buf[2];
  memset(buf, 0, 2);
  char pkt_len_buf[2];
  memset(pkt_len_buf, 0, 2);
  char pkt_buf[MAX_BUF_SIZE];
  memset(pkt_buf, 0, MAX_BUF_SIZE);


  unsigned short pkt_len=0;
  int toread = 0;
  int recv_bytes =0;
  //read first two bytes (the TCP DNS packet length);
  memset(pkt_len_buf, 0, 2);
  while(toread<2)
  {
    recv_bytes = SSL_read(myssl, (void*)buf, 1);
    if(recv_bytes<=0)
    {
      printf("receive bytes: %d\n", recv_bytes);
      return -1;
    }
    if(recv_bytes==1)
    {
      pkt_len_buf[toread]=buf[0];
      toread++;
    }
  }

  memcpy(recv_buf, pkt_len_buf, 2);
  pkt_len = ntohs(*((unsigned short*) pkt_len_buf));

  //read the rest packet;
  toread=0; 
  memset(pkt_buf, 0, MAX_BUF_SIZE);
  while(toread<pkt_len)
  {
    recv_bytes = SSL_read(myssl, (void*)buf, 1);
    if(recv_bytes<=0)
    {
      printf("receive bytes: %d\n", recv_bytes);
      return -1;
    }

    if(recv_bytes==1)
    {
      pkt_buf[toread]=buf[0];
      toread++;
    }
  }

  memcpy(recv_buf+2, pkt_buf, pkt_len);
  return pkt_len+2;
}

/*
 * create a new tcp connection 
 * */
bool create_tcp_connection(const char *host , unsigned int port, CSocket* sock)
{
  if(!sock->create(SOCK_STREAM, IPPROTO_TCP))
  {
    return false;
  }
  
  if(disable_nagle)
  {
    sock->disable_nagle();
  }
  
  //enable fast open
  if(enable_tcpfast)
  {
    sock->enable_fastopen();
  }
  //normal handshake.
  else
  {
    if(!sock->connect(host, port))
    {
      sock->close();
      return false;
    }
  }
  return true;
}

//create one ssl connection. 
SSL* build_sslconn()
{

  //no existing idle connection, need to create a new ssl connection 
  CSocket *mysock = new CSocket;
  if(!create_tcp_connection(dns_servers, dns_port, mysock)) 
  {
    fprintf(log_fp, "creat tcp connection failed\n");
    return NULL;
  }

  if(!direct_tls)
  {
    //check if the resolver support TLS or not.
    if(!resolver_tls_check(*mysock))
    {
      fprintf(log_fp, "resolver doesn't support tls\n");
      return NULL;
    }
  }

  //new ssl session;
  SSL* ssl = NULL;
  if((ssl=ssl_handshake(mysock)) == NULL)
  {
    fprintf(log_fp, "ssl handshake failed\n");
  }
  return ssl;
}

//usage information
void usage(char* prog)
{
  cout << "usage: " << prog 
  << "\n -r/--dns-resolver = resolver ip, remote t-DNS proxy ip, default 128.9.160.224"
  //<< "\n -d/--disable-nagle, disable nagle"
  //<< "\n -v/--validate-ca, validate CA"
  //<< "\n -o/--enable-tcpfast, enable TCP fastopen"
  << "\n -p/--port-number = port, remote t-DNS proxy port, default 6667"
  << "\n -l/--local-port = port, local port to listen on, default 53"
  << "\n -t/--protocol = protocol between client proxy to remote server, could be tcp or ssl"
  << "\n -e/--direct-tls, send query over tls directly, without tls negotiation. negotiation is enabled by default"
  << "\n -h/--help, show this message"
  << "\n example: sudo ./tdns_cli_proxy -t ssl -r 128.9.160.224 -p 6667 -l 53"
  << endl;
}


// parse arguments
int
options(int argc, char *argv[])
{
    static struct option longopts[]={   
          {"local-port", required_argument, NULL, 'l'}, 
          {"disable-nagle", no_argument, NULL, 'd'}, 
          {"enable-tcpfast", no_argument, NULL, 'o'}, 
          {"validate-ca", no_argument, NULL, 'v'}, 
          {"direct-tls", no_argument, NULL, 'e'}, 
          {"dns-resolver", required_argument, NULL, 'r'}, 
          {"protocol", required_argument, NULL, 't'}, 
          {"port-number", required_argument, NULL, 'p'}, 
          {"help", no_argument, NULL, 'h'}, 
          {NULL, 0, NULL, 0}
    };  

    int ch; 
    while ((ch = getopt_long(argc, argv, "r:l:p:t:dceovh", longopts, NULL)) != -1) {
        switch (ch) {
        case 't':
            proto=optarg;
            break;
        case 'd':
            disable_nagle=true;
            break;
        case 'o':
            enable_tcpfast=true;
            break;
        case 'v':
            validate_ca=true;
            break;
        case 'e':
            direct_tls=true;
            break;
        case 'p':
            dns_port=atoi(optarg);
            break;
        case 'r':
            strcpy(dns_servers, optarg);
            break;
        case 'l':
            local_port = atoi(optarg);
            break;
        case 'h':
            usage(argv[0]);
            exit(1);
        default:
            usage(argv[0]);
            exit(1);
        }   
    }   
    return optind;
}




//close TCP/TLS connection
void clean_old_connection(struct conn_hub *conn_p)
{
  //close old TCP connection.
  if(conn_p->conn_type == "tcp")
  {
    CSocket *mysock = conn_p->tcp_conn;
    mysock->close();
  }
  //close old SSL connection.
  if(conn_p->conn_type == "ssl")
  {
    SSL* myssl = conn_p->ssl_conn;
    SSL_shutdown(myssl);
    SSL_free(myssl);
  }
}

//handle ctrl+c
void sig_handler (int sig)
{
  printf("quit program....\n");
  clean_old_connection(conn_point);
  delete conn_point;
  fclose(log_fp);
  exit(0);
}


int create_conn_remote(char* remote_server, int remote_port)
{

  //tcp/ssl connection communicate with remote server. 
  CSocket *mysock = NULL;
  SSL* myssl = NULL;
  conn_point->conn_type = proto;
  if(proto=="tcp")
  {
    mysock = new CSocket;
    if(!create_tcp_connection(remote_server, remote_port, mysock)) 
    {
      fprintf(log_fp, "creat tcp connection failed to : %s:%d\n", remote_server, remote_port);
      conn_point->tcp_conn = NULL;
      return -1;
    }
    else
    {
      conn_point->tcp_conn = mysock;
    }
  }
  else if(proto=="ssl")
  {
    if(!SSL_init_done)
    {
      ERR_load_SSL_strings();
      OpenSSL_add_all_algorithms();
      SSL_library_init();
      SSL_init_done = true;
    }

    myssl = build_sslconn();
    if(myssl == NULL)
    {

      conn_point->ssl_conn = NULL;
      
      fprintf(log_fp, "failed to build a ssl connection to : %s:%d\n", remote_server, remote_port);
      fprintf(log_fp, "try tcp connection instead\n");
      //try tcp connection instead?
      mysock = new CSocket;
      if(!create_tcp_connection(remote_server, remote_port, mysock)) 
      {
        fprintf(log_fp, "creat tcp connection failed to : %s:%d\n", remote_server, remote_port);
        conn_point->tcp_conn = NULL;
        return -1;
      }
      else
      {
        conn_point->conn_type = "tcp";
        conn_point->tcp_conn = mysock;
      }
    }
    else
    {
      conn_point->ssl_conn = myssl;
    }

  }
  else
  {
    printf("protocol not supported, program exit....\n");
    exit(0);
  }


  printf("ready to communicate with: %s:%d over: %s\n", remote_server, remote_port, (conn_point->conn_type).c_str());
  return 1; 
}

int create_conn_local(int local_port)
{
  //UDP socket to communicate with local client. 
  CSocket *mysock = new CSocket;
  if(!mysock->create(SOCK_DGRAM, IPPROTO_UDP))
  {
    fprintf(log_fp, "creat UDP socket failed\n");
    exit(0);
  }

  //bind to local ip and port
  struct sockaddr_in local_addr;
  local_addr.sin_family = AF_INET;
  local_addr.sin_port = htons(local_port);
  local_addr.sin_addr.s_addr = htonl(INADDR_ANY); //inet_addr(local_server); 

  if(!mysock->bind_rawsock(local_addr))
  {
    printf("bind to port: %d failed\n", local_port);
    exit(0);
  }

  conn_point->udp_conn = mysock;
  printf("listening locally on port: %d\n", local_port);
  return 1; 
}

bool reconnection()
{
  int i=0;
  for(i=0; i< retry_count; i++)
  {
    fprintf(log_fp, "%d try to reconnect to remote server in %d seconds\n", i, retry_timeout);
    sleep(retry_timeout);
    if(create_conn_remote(dns_servers, dns_port)==1)
    {
      fprintf(log_fp, "reconnected to remote server\n");
      return true;
    }
    retry_timeout*=2;
  }

  printf("all %d tries fail\n", retry_count);
  return false;
}

struct sockaddr_in client_addr;
//thread to receive response on TCP connection
void* response_remote_to_local(void* args)
{
 
  printf("start thread to forward response from remote to local\n");
  struct conn_hub *conn_p = (struct conn_hub *)args;
  char pkt_buf[MAX_BUF_SIZE];
  char *buf=NULL;

  int recv_len = 0;
  int send_len = 0;
  while(true)
  {
    //receive response from remote server. 
    memset(pkt_buf, 0, MAX_BUF_SIZE);
    if(conn_p->conn_type == "tcp" && conn_p->tcp_conn !=NULL) 
    {
      if((recv_len=recv_tcp_response((conn_p->tcp_conn),(char*)pkt_buf, MAX_BUF_SIZE)) <0 )
      {
        fprintf(log_fp, "receive error. received:%d bytes\n", recv_len);
        clean_old_connection(conn_p);
        if(!reconnection())
        {
          cleanup();
          exit(0);
        }
        else
          continue;
      }
    }

    if(conn_p->conn_type == "ssl" && conn_p->ssl_conn !=NULL) 
    {
      if((recv_len=recv_ssl_response((conn_p->ssl_conn),(char*)pkt_buf, MAX_BUF_SIZE)) <0 )
      {
        fprintf(log_fp, "receive error. received:%d bytes\n", recv_len);
        clean_old_connection(conn_p);
        if(!reconnection())
        {
          cleanup();
          exit(0);
        }
        else
          continue;
      }
    }

    if((conn_p->conn_type == "tcp") || (conn_p->conn_type == "ssl")) 
    {
      recv_len-=2;
      buf=pkt_buf+2;
    }
    print_dns_msg((unsigned char*)buf,recv_len, log_fp);

    //forward response to local client.
    if((send_len=(conn_p->udp_conn)->send_data_UDP((char*)buf, recv_len, client_addr))<0)
    {
      fprintf(log_fp, "forward response back to client fail, pkt_len:%d, send_len:%d\n", recv_len, send_len);
    }

  }
  cleanup();
  exit(0);
}

//thread to receive response on TCP connection
void* query_local_to_remote(void* args)
{
  struct conn_hub *conn_p = (struct conn_hub *)args;
  printf("start thread to forward query from local to remote via: %s \n", conn_p->conn_type.c_str());
  char pkt_buf[MAX_BUF_SIZE];
  char buf[MAX_BUF_SIZE];

  int recv_len = 0;
  int send_len = 0;
  while((conn_p->udp_conn)->is_alive())
  {

    //receive query from local client. 
    memset(pkt_buf, 0, MAX_BUF_SIZE);
    if((recv_len=(conn_p->udp_conn)->recv_data_UDP((char*)pkt_buf, client_addr))<0)
    {
      fprintf(log_fp, "receive error. received %d bytes\n", recv_len);
    }

    print_dns_msg((unsigned char*)pkt_buf,recv_len, log_fp);
    memset(buf, 0, MAX_BUF_SIZE);
    if((conn_p->conn_type == "tcp") || (conn_p->conn_type == "ssl")) 
    {
      u_int16_t len_to_write = htons(recv_len);//this is very important
      memcpy(buf, &len_to_write, sizeof(len_to_write));
      memcpy(buf+sizeof(len_to_write), pkt_buf, recv_len);
      recv_len += sizeof(len_to_write);
    }
    else
    {
      memcpy(buf, pkt_buf, recv_len);
    }

    //forward query to remote server over TCP.
    if(conn_p->conn_type == "tcp")
    {
      if(!(conn_p->tcp_conn)->is_alive())
      {
        fprintf(log_fp, "TCP from local to remote server dead\n");
        sleep(1);
      }
      //send the query via normal TCP.
      if((send_len= (conn_p->tcp_conn)->send_data_TCP((char*)buf, recv_len))<0)
      {
        fprintf(log_fp, "TCP: forward query from client to remote server, pkt_len:%d, send_len:%d\n", recv_len, send_len);
      }

    }
    //forward query to remote server over SSL.
    if(conn_p->conn_type == "ssl")
    {
      if((send_len = SSL_write((conn_p->ssl_conn), buf, recv_len))<0)
      {
        fprintf(log_fp, "SSL: forward query from client to remote server, pkt_len:%d, send_len:%d\n", recv_len, send_len);
      }
    }

  }
  cleanup();
  exit(0);
}

void cleanup()
{
  clean_old_connection(conn_point);
  delete conn_point;
  fflush(log_fp);
  fclose(log_fp);
}

//main function
int main( int argc , char *argv[])
{
    signal(SIGQUIT, sig_handler);
    signal(SIGINT, sig_handler);


    if(argc<2)
    {
      usage(argv[0]);
      exit(1);
    }

    log_fp=fopen("/var/log/tdns_cli_proxy.log", "w+");
    if(!log_fp)
    {
      fprintf(stderr, "create log file failed, use stdout/stderr for logs\n");
      log_fp=stderr;
    }
    else
    {
      printf("log file: /var/log/tdns_cli_proxy.log\n");
    }
    //default remote dns recursive resolver and port number;
    strcpy(dns_servers, R_SERV);

    int i=0;
    int ret=0;
    //parse parameters;
    i = options(argc, argv);
    argc -= i;
    argv += i;
    
    //create TCP/TLS socket for communicating with remote resolver.
    create_conn_remote(dns_servers, dns_port);

    //create UDP socket for communicating with local client.
    create_conn_local(local_port);
    
    //one thread for receiving from remote resolver, forwarding to local client 
    ret = pthread_create(&td_remote_to_local, NULL, response_remote_to_local, static_cast<void*>(conn_point));
    if(ret!=0)
      fprintf(log_fp, "create thread failed\n");

    //one thread for receiving from local client, forwarding to remote resolver.
    ret = pthread_create(&td_local_to_remote, NULL, query_local_to_remote, static_cast<void*>(conn_point));
    if(ret!=0)
      fprintf(log_fp, "create thread failed\n");

    pthread_join(td_remote_to_local,NULL);
    pthread_join(td_local_to_remote,NULL);

    cleanup();
    return 0;
}

