/*
 * Copyright (C) 2013 by the University of Southern California
 * $Id: utility.cc 19925 Tue Feb 11 14:41:03 PST 2014 Zi Hu $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "utility.h"
#include "types.h"
#include <bind/resolv.h>
#include <bind/arpa/nameser.h>

//split a string to an vector;
void Splitstring(char* cStr, char* cDelim, vector<string> &sItemVec)
{
	char* p;
	p=strtok(cStr, cDelim);
	while (p!=NULL)
	{
		sItemVec.push_back(p);
		p=strtok(NULL, cDelim);
	}
}


const std::string get_current_time() 
{
    time_t     now = time(0);
    struct tm  tstruct;
    char       buf[80];
    tstruct = *localtime(&now);
    // Visit http://en.cppreference.com/w/cpp/chrono/c/strftime
    // for more information about date/time format
    strftime(buf, sizeof(buf), "%Y-%m-%d.%X", &tstruct);
    return buf;
}

//compute median of an array
double compute_median(vector<unsigned long>& arr)
{
  double median;
  size_t size = arr.size();

  sort(arr.begin(), arr.end());

  if (size  % 2 == 0)
  {
      median = (arr[size / 2 - 1] + arr[size / 2]) / 2;
  }
  else 
  {
      median = arr[size / 2];
  }
  return median; 
}

//compute the mean of an array
double compute_mean(vector<unsigned long>& arr)
{
    size_t size = arr.size();
    double sum = accumulate(arr.begin(), arr.end(), 0.0);
    double mean = sum / size;
    return mean;
}

 
/*
 * This will convert www.google.com to 3www6google3com 
 * */
//function borrowed from "http://www.binarytides.com/dns-query-code-in-c-with-linux-sockets/", Author : Silver Moon
void ChangetoDnsNameFormat(unsigned char* dns,unsigned char* host) 
{
    int lock = 0 , i;
    unsigned char tmp[MAX_BUF_SIZE];
    memset(tmp, 0, MAX_BUF_SIZE);
    strcpy((char*)tmp,(char*)host);
    strcat((char*)tmp,".");
     
    for(i = 0 ; i < strlen((char*)tmp) ; i++) 
    {
        if(tmp[i]=='.') 
        {
            *dns++ = i-lock;
            for(;lock<i;lock++) 
            {
                *dns++=tmp[lock];
            }
            lock++; //or lock=i+1;
        }
    }
    *dns++='\0';
}

unsigned short construct_dns_query(unsigned char* buf, int buf_size, const unsigned char *qname, int qclass, int qtype, bool is_tcp)
{
  /*
  if (!done_res_init) {
    void res_init();
    done_res_init = true;
  }
  */

  void res_init();
  _res.options |= RES_RECURSE;

  //int res_mkquery(int op, const char *dname, int class, int type, const u_char *data, 
  //    int datalen, const u_char *newrr,u_char *buf, int buflen)
  unsigned char* buf_pos = buf;
  if (is_tcp) {
    buf_pos = buf + 2;
    buf_size -= 2;
  }   
  int query_len = res_mkquery(
    ns_o_query, //regular query
    (char*)qname,
    qclass, //class
    qtype,  //type
    NULL,   //A buffer containing the data for inverse queries. It is NULL when op is QUERY.
    0,    //length of pre-one
    NULL,   //A buffer used for the dynamic update code
    buf_pos,    // buffer for the query
    buf_size);  // size of the buffer
  if (query_len == -1) {
    herror("res_mkquery");
    exit(-1);
  }

  if (is_tcp) {
    u_int16_t len_to_write = htons(query_len);//this is very important
    memcpy(buf, &len_to_write, sizeof(len_to_write));
    query_len += sizeof(len_to_write);
  } else {
  }

  return query_len;
}

//build the edns part in the query
int build_edns (res_state statp, int n0, unsigned char *buf, int buf_len, int anslen, bool tls_flag)
{
  //int res_nopt(res_state statp,
  //   int n0,    /*%< current offset in buffer */
  //   u_char *buf,   /*%< buffer to put query */
  //   int buflen,    /*%< size of buffer */
  //   int anslen)    /*%< UDP answer buffer size */
  anslen > 0 ? : anslen = UDP_ANS_SIZE; 
  int n = res_nopt(statp, n0, buf, buf_len, anslen);
  unsigned char *rdata = &buf[n];
  if (n > 0 && (statp->options & RES_NSID) != 0U) {
    n = res_nopt_rdata(statp, n, buf, sizeof(buf), rdata, NS_OPT_NSID, 0, NULL);
  }
  if (tls_flag) { //we will set the second bit of Z filed to indicate "enabling TLS"
      //the most significant bit of the Z field has been used for DNSSEC ["DNSSEC OK" (DO) bit]
    unsigned char * zp = buf + n0 + ZBIT_OFFSET - 1;
    u_int16_t z_old = ns_get16(zp);
    u_int16_t z_new = z_old;
    z_new |= DNS_MESSAGEEXTFLAG_TLS;
    ns_put16(z_new, zp);
  }   
  return n;
}


unsigned short construct_dns_query_EDNS(unsigned char* buf, int buf_size, const char *qname, int qclass, int qtype, bool is_tcp)
{
  /*
  if (!done_res_init) {
    void res_init();
    done_res_init = true;
  }
  */

  //void res_init();
  //_res.options |= RES_RECURSE;
   
  res_state res_udef;
  res_udef = __res_state();//user defines res, it's better not use the internal _res
  res_udef->options |= RES_RECURSE;//cout << res_udef->options<<endl;
  if (res_ninit(res_udef) == -1 ) {
    fprintf(stderr, "res_ninit() fails!");
    exit(1);
  }


  unsigned char* buf_pos = buf;
  if (is_tcp) {
    buf_pos = buf + 2;
    buf_size -= 2;
  }   

  int query_len = res_nmkquery(res_udef, ns_o_query, qname, qclass, qtype, NULL, 0, NULL, buf_pos, buf_size);
  if (query_len == -1) {
    herror("res_nmkquery");
    exit(-1);
  }

  if (query_len > 0) {
    query_len = build_edns (res_udef, query_len, buf_pos, buf_size, UDP_ANS_SIZE, true);
  }

  if (is_tcp) {
    u_int16_t len_to_write = htons(query_len);//this is very important
    memcpy(buf, &len_to_write, sizeof(len_to_write));
    query_len += sizeof(len_to_write);
  } else {
  }
  return query_len;
}

     

//format a DNS query
//function borrowed from "http://www.binarytides.com/dns-query-code-in-c-with-linux-sockets/", Author : Silver Moon
unsigned short format_dns_query(char* buf, unsigned short &qlen, unsigned char* host, int query_type, int type, unsigned short msgid)
{
    unsigned char *qname,*reader;
    struct DNS_HEADER *dns = NULL;
    struct QUESTION *qinfo = NULL;
 
    //Set the DNS structure to standard queries
    unsigned short dh_len=0;
    memset(buf, 0, MAX_BUF_SIZE);
    unsigned char* buf_init_pos = (unsigned char*)buf;

    if(type==QT_TCP)
    {
      buf_init_pos += sizeof(dh_len);
    }

    dns = (struct DNS_HEADER *)buf_init_pos;
    if(msgid==0)
      dns->id = (unsigned short) htons(getpid());
    else
      dns->id = htons(msgid);
    dns->qr = 0; //This is a query
    dns->opcode = 0; //This is a standard query
    dns->aa = 0; //Not Authoritative
    dns->tc = 0; //This message is not truncated
    dns->rd = 1; //Recursion Desired
    dns->ra = 0; //Recursion not available
    dns->z = 0;
    dns->ad = 0;
    dns->cd = 0;
    dns->rcode = 0;
    dns->q_count = htons(1); //we have only 1 question
    dns->ans_count = 0;
    dns->auth_count = 0;
    dns->add_count = 0;
 
    //point to the query portion
    qname =(unsigned char*)(buf_init_pos+sizeof(struct DNS_HEADER));
 
    ChangetoDnsNameFormat(qname , host);
    qinfo =(struct QUESTION*)(buf_init_pos+sizeof(struct DNS_HEADER) + (strlen((const char*)qname) + 1)); //fill it
 
    qinfo->qtype = htons( query_type ); //type of the query , A , MX , CNAME , NS etc
    qinfo->qclass = htons(1); //its internet

    qlen=strlen((char*)qname)+1;
    //get the length of the packet;
    unsigned short len=sizeof(struct DNS_HEADER) + (strlen((const char*)qname)+1) + sizeof(struct QUESTION);
    if(type==QT_TCP)
    {
      dh_len = htons(len);
      memcpy(buf, &dh_len, sizeof(dh_len));
      return len+sizeof(dh_len);
    }
    else
    {
      return len;
    }
}


 

/*
 * compute checksume
 */
unsigned short in_cksum(unsigned short *addr, int len)
{
    register int sum = 0;
    u_short answer = 0;
    register u_short *w = addr;
    register int nleft = len;
	  
    while (nleft > 1)
    {
	sum += *w++;
	nleft -= 2;
    }
    /* mop up an odd byte, if necessary */
    if (nleft == 1)
    {
	*(u_char *) (&answer) = *(u_char *) w;
	sum += answer;
    }
    /* add back carry outs from top 16 bits to low 16 bits */
    sum = (sum >> 16) + (sum & 0xffff);       /* add hi 16 to low 16 */
    sum += (sum >> 16);               /* add carry */
    answer = ~sum;              /* truncate to 16 bits */
    return (answer);
}

unsigned short in_cksum_tcp(int src, int dst, unsigned short *tcp_packet, int len)
{
    //len = sizeof(struct tcphdr) + tcp_opt_len + tcp_payload_len;
    struct psd_tcp psd_tcph;
    memset(&psd_tcph, 0, sizeof(struct psd_tcp));

    psd_tcph.src.s_addr = src;
    psd_tcph.dst.s_addr = dst;
    psd_tcph.pad = 0;
    psd_tcph.proto = IPPROTO_TCP;
    psd_tcph.tcp_len = htons(len);

    //buf len = psudo header len + tcp packet len;
    int tcp_cksum_buf_len = sizeof(struct psd_tcp) + len;
    char* buf = new char [tcp_cksum_buf_len];
    memset(buf, 0, tcp_cksum_buf_len);

    memcpy(buf, &psd_tcph, sizeof(struct psd_tcp));
    memcpy(buf + sizeof(struct psd_tcp), (unsigned char*)tcp_packet, len);

    u_short ans;
    ans = in_cksum((unsigned short *)buf, tcp_cksum_buf_len);
    delete [] buf;
    return (ans);
}

void print_section(ns_msg handle, int count, ns_sect sec_t, const char* sec_name, FILE* log_fp)
{
  ns_rr rr;
  char buf[MAX_BUF_SIZE];
  //cout << sec_name << ":" << endl;
  for (int rrnum=0; rrnum<count; rrnum++) {
    memset(&rr, 0, sizeof(rr));
    if (ns_parserr(&handle, sec_t, rrnum, &rr)) {
      cerr<<"[ERROR]ns_parserr"<<endl;
      exit(-1);
    }
    memset(buf, '\0', sizeof(buf));
    ns_sprintrr (&handle, &rr, NULL, NULL, buf, sizeof (buf));
    //cout << buf << endl;
    fprintf(log_fp, "%s\n", buf);
  }
}

unsigned short get_dns_msg_id(unsigned char* msg, int msg_len)
{
  ns_msg  handle;
  if (ns_initparse(msg, msg_len, &handle) < 0) {
    perror("ns_initparse");
    exit(-1);
  }

  u_int16_t ID      = ns_msg_id(handle);
  u_int16_t QR      = ns_msg_getflag(handle, ns_f_qr);
  u_int16_t AA      = ns_msg_getflag(handle, ns_f_aa);
  u_int16_t TC      = ns_msg_getflag(handle, ns_f_tc);
  u_int16_t RD      = ns_msg_getflag(handle, ns_f_rd);
  u_int16_t RA      = ns_msg_getflag(handle, ns_f_ra);
  u_int16_t Z       = ns_msg_getflag(handle, ns_f_z);
  u_int16_t Opcode  = ns_msg_getflag(handle, ns_f_opcode);
  u_int16_t Rcode   = ns_msg_getflag(handle, ns_f_rcode);
  u_int16_t QdCount = ns_msg_count(handle, ns_s_qd);
  u_int16_t AnCount = ns_msg_count(handle, ns_s_an);
  u_int16_t NsCount = ns_msg_count(handle, ns_s_ns);
  u_int16_t ArCount = ns_msg_count(handle, ns_s_ar);
  return ID;
}


void print_dns_msg(unsigned char* msg, int msg_len, FILE* log_fp)
{
  ns_msg  handle;
  if (ns_initparse(msg, msg_len, &handle) < 0) {
    perror("ns_initparse");
    exit(-1);
  }

  u_int16_t ID      = ns_msg_id(handle);
  u_int16_t QR      = ns_msg_getflag(handle, ns_f_qr);
  u_int16_t AA      = ns_msg_getflag(handle, ns_f_aa);
  u_int16_t TC      = ns_msg_getflag(handle, ns_f_tc);
  u_int16_t RD      = ns_msg_getflag(handle, ns_f_rd);
  u_int16_t RA      = ns_msg_getflag(handle, ns_f_ra);
  u_int16_t Z       = ns_msg_getflag(handle, ns_f_z);
  u_int16_t Opcode  = ns_msg_getflag(handle, ns_f_opcode);
  u_int16_t Rcode   = ns_msg_getflag(handle, ns_f_rcode);
  u_int16_t QdCount = ns_msg_count(handle, ns_s_qd);
  u_int16_t AnCount = ns_msg_count(handle, ns_s_an);
  u_int16_t NsCount = ns_msg_count(handle, ns_s_ns);
  u_int16_t ArCount = ns_msg_count(handle, ns_s_ar);

  fprintf(log_fp, "******time: %s******\n", (get_current_time()).c_str());
  fprintf(log_fp, "%s, len: %d, id:%d\n", (QR==0 ? "query" : "response"), msg_len, ID);

  print_section(handle, QdCount, ns_s_qd, "Que", log_fp);
  //print_section(handle, AnCount, ns_s_an, "Ans", log_fp);
  //print_section(handle, NsCount, ns_s_ns, "Aus", log_fp);
  //print_section(handle, ArCount, ns_s_ar, "Add", log_fp);
  fflush(log_fp);
}

bool read_dnslist(char* fn, vector<string>& dns_list)
{
  char buf[MAX_BUF_SIZE];
  FILE*fp = fopen(fn, "r");
  if(!fp)
    return false;
  while (fgets(buf, MAX_BUF_SIZE, fp))
  {
    buf[strlen(buf)-1]='\0';
    if(buf[0] == '#')
      continue;
    dns_list.push_back(buf);
    memset(buf, 0, MAX_BUF_SIZE);
  }
  fclose(fp);
  return true;
}

void generate_uniq_random(vector<string> &dns_list, int num)
{
  srand (time(NULL));
  set<string> uniq_dns;
  if(num>dns_list.size())
  {
    printf("exceed max: %d\t%lu\n", num, dns_list.size());
  }

  int i=0;
  for(i=0; i<num; )
  {
    int rd_index = rand()%(dns_list.size());
    if(uniq_dns.find(dns_list[rd_index])!= uniq_dns.end())
    {
      continue;
    }
    else
    {
      printf("%s\n", (dns_list[rd_index]).c_str());
      uniq_dns.insert(dns_list[rd_index]);
      i++;
    }
  }
}

bool is_tls_set (unsigned char* msg, int msg_len)
{
  //LOG( LOG_DEBUG, "[DEBUG] is_tls_set()\n");
  ns_msg  handle;
  if (ns_initparse(msg, msg_len, &handle) < 0) {
    perror("ns_initparse");
    exit(-1);
  }
  u_int16_t ArCount = ns_msg_count(handle, ns_s_ar);
  if (ArCount == 0)//no OPT
    return false;

  ns_rr rr;

  //find the first query name: this should be '.' 
  memset(&rr, 0, sizeof(rr));
  if (ns_parserr(&handle, ns_s_qd, 0, &rr)) {
    perror("ns_parserr");
  }
  string rr_name = ns_rr_name(rr);
  //LOG( LOG_DEBUG, "[DEBUG] is_tls_set(): query name is %s\n", rr_name.c_str());
  int q_type = ns_rr_type(rr);
  int q_class = ns_rr_class(rr);
  fprintf(stderr, "TYPE :%d, CLASS:%d\n", q_type, q_class);

  //EDNS with tls bit set?
  u_int16_t rr_type = 0;
  u_int32_t rr_ttl = 0;
  for (int rrnum=0; rrnum<ArCount; rrnum++) {
    memset(&rr, 0, sizeof(rr));
    if (ns_parserr(&handle, ns_s_ar, rrnum, &rr)) {
      perror("[ERROR] ns_parserr");
    }
    //ns_sprintrr (&handle, &rr, NULL, NULL, buf, sizeof (buf));
    rr_type = ns_rr_type(rr);
    rr_ttl = ns_rr_ttl(rr);
    printf("[DEBUG] TTL bits:0x%08x\n",rr_ttl);
    if (rr_type == ns_t_opt) {//OPT, ENDS
      if (rr_ttl & DNS_MESSAGEEXTFLAG_TLS) {//tls bit is set
        //LOG( LOG_DEBUG, "[DEBUG] tls bit is set\n");
        //LOG( LOG_DEBUG, "[DEBUG] TTL bits:0x%08x\n",rr_ttl);
        return true;
      }
    }
  }
  return false;
}

